{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<h1>Computation of $\\mathcal{GT}_1(G)$</h1>\n",
    "\n",
    "<em>Pierre Guillot</em>\n",
    "\n",
    "This computes the group $\\mathcal{GT}_1(G)$, described in my paper <em>The Grothendieck-Teichmüller group of a finite group and $G$-dessins d'enfants</em>. \n",
    "\n",
    "In order to use and modify the script, you first need to:\n",
    "<ul><li>create an accound on SageMathCloud, if you haven't got one already, </li>\n",
    "<li>create at least one \"project\" on SageMathCloud.</li></ul>\n",
    "\n",
    "Then you can click on \"Copy Jupyter Notebook to your project...\", in the top left hand corner.\n",
    "\n",
    "Alternatively, if you have a recent version of Sage installed, with the Jupyter Notebook working, you can click on \"Download\"."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": [
    "IsContained= libgap.function_factory(\"\\in\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": [
    "class GT1:\n",
    "    \n",
    "    def __init__(self, G, verbose= False, very_verbose= False):\n",
    "        \n",
    "        self.G= G\n",
    "        self.verbose= verbose or very_verbose\n",
    "        self.very_verbose= very_verbose\n",
    "        \n",
    "    def compute_cover_group(self):\n",
    "        \n",
    "        if self.verbose:\n",
    "            print \"*** computing cover\"\n",
    "        \n",
    "        ##product group\n",
    "        GG= self.G.DirectProduct(self.G)\n",
    "        i1= GG.Embedding(1)\n",
    "        i2= GG.Embedding(2)\n",
    "        p1= GG.Projection(1)\n",
    "        p2= GG.Projection(2)\n",
    "\n",
    "        ##automorphism group of G\n",
    "        autG= self.G.AutomorphismGroup()\n",
    "        #convert the elements of autG into automorphisms of GG\n",
    "        gensautG= autG.GeneratorsOfGroup()\n",
    "        gensGG= GG.GeneratorsOfGroup()\n",
    "        imgs= []\n",
    "        for phi in gensautG:\n",
    "            phi_on_gens= [ (p1*phi*i1).Image(gh) * (p2*phi*i2).Image(gh) for gh in gensGG  ]\n",
    "            phiGG= GG.GroupHomomorphismByImages(GG, gensGG, phi_on_gens)\n",
    "            imgs.append(phiGG)\n",
    "        #using this, embed autG in autGG...\n",
    "        diagautG= libgap.Group(imgs)\n",
    "\n",
    "        ### classes of pairs in G\n",
    "        classes= diagautG.OrbitsDomain(GG)\n",
    "\n",
    "        if self.very_verbose:\n",
    "            print len(classes), \"classes of pairs\"\n",
    "\n",
    "        def is_generating_pair(C):\n",
    "            pair= C[0]\n",
    "            x=  p1.Image(pair)\n",
    "            y=  p2.Image(pair)\n",
    "\n",
    "            return self.G.Subgroup([x, y]) == self.G\n",
    "\n",
    "        gens_classes= [ C for C in classes if is_generating_pair(C) ]\n",
    "        r= len(gens_classes)\n",
    "        \n",
    "        if self.very_verbose:\n",
    "            print r,\"of them generate G\"\n",
    "\n",
    "        ###build G^r and then GB\n",
    "        Gr= libgap.DirectProduct([ self.G for i in range(r) ])\n",
    "        inj= [ Gr.Embedding(i) for i in [1..r] ]\n",
    "        self.x= prod([ (p1*inj[i]).Image(gens_classes[i][0])  for i in range(r)  ])\n",
    "        self.y= prod([ (p2*inj[i]).Image(gens_classes[i][0])  for i in range(r)  ])\n",
    "        self.GB= Gr.Subgroup([self.x, self.y])\n",
    "\n",
    "\n",
    "    def compute_automorphisms(self):\n",
    "        if self.verbose:\n",
    "            print \"*** computing representatives for GT1\"\n",
    "            \n",
    "        # for convenience\n",
    "        GB= self.GB\n",
    "        x= self.x\n",
    "        y= self.y \n",
    "        one= GB.Identity()\n",
    "        \n",
    "        ### the centralizers \n",
    "        Cx= GB.Centralizer(x)\n",
    "        Cy= GB.Centralizer(y)\n",
    "\n",
    "        ### find the double cosets\n",
    "        doubles= GB.DoubleCosets(Cx, Cy)\n",
    "        if self.very_verbose:\n",
    "            print \"found\", len(doubles), \"double cosets\"\n",
    "        wait= len(doubles)\n",
    "        \n",
    "        ### prepare theta and delta\n",
    "        theta= GB.GroupHomomorphismByImages(GB, [x, y], [y, x])\n",
    "        delta= GB.GroupHomomorphismByImages(GB, [x, y], [y^(-1)*x^(-1), y])\n",
    "\n",
    "        ### do all the tests !\n",
    "        \n",
    "        self.auts= []\n",
    "        \n",
    "        for i, C in enumerate(doubles):\n",
    "            if self.very_verbose:\n",
    "                print \"coset\", i+1, \"out of\", wait\n",
    "            #take a representative\n",
    "            c= C.Representative()\n",
    "            xp= x^c\n",
    "            \n",
    "            #first test: <x^c, y> is all of GB\n",
    "            GBp= GB.Subgroup([xp, y])\n",
    "            if GBp != GB:\n",
    "                continue # try next C\n",
    "       \n",
    "            #second test: commutation with theta\n",
    "            ct= c^theta\n",
    "            if not IsContained(one, Cx.DoubleCoset(c*ct, Cy^ct)):\n",
    "                continue\n",
    "     \n",
    "            #third test: automorphism is actually well-defined\n",
    "            phi= GB.GroupHomomorphismByImages(GB, [x, y], [xp, y])\n",
    "            if phi == libgap.eval(\"fail\"):\n",
    "                continue\n",
    "\n",
    "            #fourth test: commutation with delta\n",
    "            if not libgap.IsInnerAutomorphism( phi*delta*phi^(-1)*delta^(-1)):\n",
    "                continue\n",
    "            \n",
    "            #if we're here, all the tests have been passed!\n",
    "            if self.very_verbose:\n",
    "                print \"found an element in GT1(G) !\"\n",
    "                \n",
    "            self.auts.append(phi)\n",
    "\n",
    "  "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": [
    "#G= libgap.SmallGroup(64,34)\n",
    "G= libgap.PSL(2,2)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "\"S3\""
      ]
     },
     "execution_count": 5,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "which_group(G)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": [
    "gt= GT1(G, very_verbose= True)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "*** computing cover\n",
      "11 classes of pairs\n",
      "3 of them generate G\n"
     ]
    }
   ],
   "source": [
    "gt.compute_cover_group()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "*** computing representatives for GT1\n",
      "found 3 double cosets\n",
      "coset 1 out of 3\n",
      "found an element in GT1(G) !\n",
      "coset 2 out of 3\n",
      "found an element in GT1(G) !\n",
      "coset 3 out of 3\n"
     ]
    }
   ],
   "source": [
    "gt.compute_automorphisms()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "[[ (2,3)(5,6)(7,8,9), (1,2)(4,5,6)(8,9) ] -> [ (2,3)(5,6)(7,8,9), (1,2)(4,5,6)(8,9) ],\n",
       " [ (2,3)(5,6)(7,8,9), (1,2)(4,5,6)(8,9) ] -> [ (2,3)(5,6)(7,9,8), (1,2)(4,5,6)(8,9) ]]"
      ]
     },
     "execution_count": 9,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "gt.auts"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "SageMath (stable)",
   "language": "",
   "name": "sagemath"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 2
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython2",
   "version": "2.7.13"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 1
}
